import fetch from 'node-fetch'
import crypto from 'crypto'
import { config } from './config.js'
import promClient from 'prom-client'
import { logger } from './logger.js'
import { isJSFile, viteToOxManifest } from './util.js'

async function fetchData (path, baseUrl, appendix) {
  const response = await fetch(new URL(path, baseUrl))
  if (!response.ok) throw new Error(`Error fetching file: ${path}`)
  const resBuffer = await response.arrayBuffer()
  const appendixLength = appendix?.length || 0
  const content = Buffer.alloc(resBuffer.byteLength + appendixLength)

  content.fill(Buffer.from(resBuffer), 0, resBuffer.byteLength)
  if (appendix) content.write(appendix, resBuffer.byteLength)

  const sha256Sum = crypto.createHash('sha256').update(content).digest('base64')
  return [path, {
    'content-type': response.headers.get('content-type'),
    sha256Sum,
    content
  }]
}

const fileCounter = new promClient.Counter({
  name: 'core_ui_middleware_file_cache_fetches',
  help: 'Number of fetched files'
})
const fileErrorCounter = new promClient.Counter({
  name: 'core_ui_middleware_file_cache_fetch_errors',
  help: 'Number of errors while fetching files'
})

const pullThroughRequestCounter = new promClient.Counter({
  name: 'core_ui_middleware_file_cache_pull_through_request_count',
  help: 'Number of files requested after warmup (not referenced in manifest)'
})

const pullThroughCacheCounter = new promClient.Counter({
  name: 'core_ui_middleware_file_cache_pull_through_cache_size',
  help: 'Number of files added after warmup (not referenced in manifest)'
})

class FileCache {
  constructor () {
    this._cache = {}
    this._manifests = {}
    this._hash = ''
    this._dependencies = {}
    this._oxManifests = {}
    this._isCaching = false
  }

  async warmUp (manifests, deps) {
    if (this._hash === manifests.__hash__) return logger.debug(`Tried to warm up the filecache with the same version: ${manifests.__hash__}`)
    if (this._isCaching) return logger.debug('Cache.warmup is already running')
    this._isCaching = true

    logger.debug('beginning to warm up cache')
    pullThroughCacheCounter.reset()
    pullThroughRequestCounter.reset()
    const cache = Object.fromEntries(await (async function () {
      const files = Object.keys(deps)
      const chunkSize = 50
      const result = []
      while (files.length > 0) {
        result.push.apply(result, (await Promise.all(files.splice(0, chunkSize).map(async file => {
          try {
            const manifest = manifests[file] || Object.values(manifests).find(m =>
              m.file === file ||
              (m?.assets?.indexOf(file) >= 0) ||
              (m?.css?.indexOf(file) >= 0)
            )
            if (!manifest) {
              logger.error(`could not find manifest for "${file}"`)
              return null
            }
            let appendix
            if (manifest.css && isJSFile(file)) {
              const cssString = manifest.css.map(file => `"${file}"`).join(',')
              appendix = `\n/*injected by ui-middleware*/document.dispatchEvent(new CustomEvent("load-css",{detail:{css:[${cssString}]}}))`
            }
            return await fetchData(file, manifest.meta.baseUrl, appendix)
          } catch (e) {
            fileErrorCounter.inc()
            logger.error(e)
          }
        }))).filter(data => Array.isArray(data) && data.length === 2))
      }
      fileCounter.inc(result.length)
      return result
    }()))

    this._cache = cache
    this._manifests = manifests
    this._hash = manifests.__hash__
    this._dependencies = deps
    this._oxManifests = viteToOxManifest(manifests)
    this._isCaching = false

    logger.debug('cache warmed up')
  }

  async fetchAndStore (path) {
    if (config.urls.length === 0) await config.load()
    pullThroughRequestCounter.inc()
    const [[key, value]] =
      (await Promise.allSettled(config.urls.map(baseUrl => fetchData(path, baseUrl))))
        .filter(r => r.status === 'fulfilled').map(r => r.value)
    this._cache[key.slice(1)] = value
    pullThroughCacheCounter.inc()
    return value
  }

  get (path) {
    return this?._cache[path.slice(1)] || {}
  }

  get manifests () {
    return this?._manifests
  }

  get hash () {
    return this?._hash
  }

  get dependencies () {
    return this?._dependencies
  }

  get oxManifests () {
    return this?._oxManifests
  }
}

export const fileCache = new FileCache()
